#!/usr/bin/perl
#This is used to produce a negative dataset of genomic islands. In other words, it finds regions conserved across multiple related genomes above a certain size.
#This runs mauve across all genomes that have had their genome distances calculated using cvtree
#Uses tables in the microbedb database; including the replicon_distance table which should have been previously filled using the run_cvtree.pl script
use strict;
use warnings;

use lib "perl_lib";
use lib "microbedb/MicrobeDB";

#local modules
use Mauve;
use Buster;
use Genome_Picker;

#from the microbe API
use Search;

my $microbedb_version = 1;

my $root_folder = "./";

#Only use this if running on Buster (a cluster of machines hosted by the Brinkman Lab)
my $parallel = 0;

#These cutoff variables can be adjusted before starting
#################################################

#Maximum distance needed between query genome and comparative genome (cvtree distance)
our $MAX_CUTOFF = 0.42;

#Minimum distance required between EVERY genome to be used
#Increasing this will result in a set of genomes that are more spread out
#However, increasing this too much will limit the number of query genomes that can be used for GI prediction
our $MIN_CUTOFF = 0.10;

#genomes that have less than this number of genomes with distance between min and max cutoff will not be run
our $MIN_COMPARE_CUTOFF = 3;

#Setting a maxmimum distance that at least one genome has to be within
#This is to ensure that the POSITIVE DATASET is searching for events on a similar time scale
#Decreasing this will reduce our positive dataset
our $MAX_DIST_SINGLE_CUTOFF = 0.30;

#Setting a minimum distance that at least one genome has to be over
#This is to ensure that the NEGATIVE DATASET is finding old enough conserved regions
#Increasing this will reduce our negative dataset
our $MIN_DIST_SINGLE_CUTOFF = 0.34;

#Parameters for Mauve backbone output
our $MIN_BACKBONE_SIZE = 8000;

#usually use 400 for max gap size
our $MAX_GAP_SIZE = 300;
################################################

#Make a search and replicon object to get a list of all chromosomes from a particular version
my $search_obj = new Search();
my $rep_obj    =
  new Replicon( version_id => $microbedb_version, rep_type => 'chromosome' );

my @result_objs = $search_obj->object_search($rep_obj);

my $total_jobs = scalar(@result_objs);
print "Total replicons to process: $total_jobs\n";

my $results_folder = $root_folder . "/negative";

my $folder = $results_folder;

#make the results directory
`mkdir -p $folder`;

my $picker_obj = new Genome_Picker(
    max_cutoff             => $MAX_CUTOFF,
    min_cutoff             => $MIN_CUTOFF,
    min_compare_cutoff     => $MIN_COMPARE_CUTOFF,
    max_dist_single_cutoff => $MAX_DIST_SINGLE_CUTOFF,
    min_dist_single_cutoff => $MIN_DIST_SINGLE_CUTOFF,
    microbedb_version      => $microbedb_version,
);
print "Using the following cutoffs for picking genomes:\n";

#print out the cutoffs for picking genomes
$picker_obj->print_object();

my @prepared_genomes;

#loop through each chromosome and find out what genomes can be used for comparision
foreach my $curr_rep_obj (@result_objs) {

    #Find/Calculate the the genomes we should use in our comparisions
    my @good_genomes = $picker_obj->find_comparative_genomes($curr_rep_obj);
    if (@good_genomes) {
        push( @prepared_genomes, [ $curr_rep_obj, \@good_genomes ] );
    }

}
my $num_good_queries = scalar(@prepared_genomes);
print "Number of replicons that passed the screening: $num_good_queries\n";

#Uncomment the following if you only want to see what genomes will be run
#exit;

my $job_count = 0;

#Make buster object and check all nodes on buster
my $buster;
if ($parallel) {
    $buster = new Buster();
}

foreach (@prepared_genomes) {
    my $query_replicon     = $_->[0];
    my @comparison_genomes = @{ $_->[1] };

    #		my $number = scalar(@comparison_genomes);
    #		print "$temp\t$number \n";
    $job_count++;

    if ($parallel) {

        #Get a free node on buster or wait until there is a free one
        my $node = $buster->get_node();

        #Use a fork command here so that it uses more nodes on buster
        if ( my $pid = fork() ) {

     #child of the fork
     #wait a couple of seconds before moving to the next iteration in the loop
            sleep(2);
        } else {

            my $name = $query_replicon->rpv_id;
            print "$job_count Running rpv_id: $name on node: $node\n";

            my $mauve_results_obj =
              do_mauve_alignment( $node, $query_replicon,
                @comparison_genomes );

            #do some stuff here
            my $backbone_results = $mauve_results_obj->backbone_output();
            my @backbones;
            my $total = 0;
            open( my $MYBACKBONE, '>', "$folder/$name.mybackbone" )
              || die "Can't open $name.mybackbone for writing: $!";
            foreach (@$backbone_results) {
                if ( $_ =~ /(\d+)\t(\d+)/ ) {
                    push( @backbones, [ $1, $2 ] );
                    my $size = $2 - $1;
                    $total += $size;
                    print $MYBACKBONE "$1\t$2\t$size\n";
                }
            }
            close($MYBACKBONE);

            my $rep_size       = $query_replicon->rep_size();
            my $percent_genome =
              sprintf( "%.2f", ( $total / $rep_size ) * 100 );

            #print "$name\t$total\t$percent_genome\n";

#print "ID: $name Found: $count Total: $total\n";
#open(ALIGN,">$folder/$name.mauve");
#print ALIGN @{$mauve_results_obj->output()} if defined($mauve_results_obj->output()) ;
#close(ALIGN);

            #open( my $BACKBONE,'>',"$folder/$name.backbone" );
            #print $BACKBONE @$backbone_results if defined($backbone_results);
            #close($BACKBONE);

            #open( my $XMFA, '>', "$folder/$name.xmfa" );
            #print $XMFA @{ $mauve_results_obj->output_alignment() }
            #  if defined( $mauve_results_obj->output_alignment() );
            #close($XMFA);
            exit;
        }
    } else {

        my $name = $query_replicon->rpv_id;
        print "$job_count Running rpv_id: $name \n";

        my $mauve_results_obj =
          do_mauve_alignment( 0, $query_replicon, @comparison_genomes );

        #do some stuff here
        my $backbone_results = $mauve_results_obj->backbone_output();
        my @backbones;
        my $total = 0;
        open( my $MYBACKBONE, '>', "$folder/$name.mybackbone" )
          || die "Can't open $name.mybackbone for writing: $!";
        foreach (@$backbone_results) {
            if ( $_ =~ /(\d+)\t(\d+)/ ) {
                push( @backbones, [ $1, $2 ] );
                my $size = $2 - $1;
                $total += $size;
                print $MYBACKBONE "$1\t$2\t$size\n";
            }
        }
        close($MYBACKBONE);

        my $rep_size       = $query_replicon->rep_size();
        my $percent_genome = sprintf( "%.2f", ( $total / $rep_size ) * 100 );
    }
}


sub do_mauve_alignment {
    my ( $node, @genomes ) = @_;

    my $mauve_obj;
    if ( $node eq 0 ) {
        $mauve_obj = new Mauve(
            'backbone_size'    => $MIN_BACKBONE_SIZE,
            'max_backbone_gap' => $MAX_GAP_SIZE,
            buster             => 0,
            'quiet'            => 1,
        );
    } else {
        $mauve_obj = new Mauve(
            'backbone_size'    => $MIN_BACKBONE_SIZE,
            'max_backbone_gap' => $MAX_GAP_SIZE,
            buster             => 1,
            buster_node        => $node,
            'quiet'            => 1,
        );
    }
    my @seqs;
    my $search_obj = new Search( return_obj => 'GenomeProject' );
    foreach my $curr_genome (@genomes) {

        #need the paths to the genome projects
        my ($gpo) = $search_obj->object_search($curr_genome);

        #produce the path to the proper file
        my $seq = $gpo->gpv_directory() . $curr_genome->file_name() . ".fna";
        push( @seqs, $seq );
    }

    #print "Running: $seqs[0] \n";
    return $mauve_obj->run(@seqs);
}

