package Mauve;

#Wraps the multiple alignment program, Mauve, into a perl object that is runnable from within a perl script
use strict;
use warnings;
our $AUTOLOAD;    # before Perl 5.6.0 say "use vars '$AUTOLOAD';"
use Carp;
use File::Temp qw/ :mktemp /;
use Mauve_Results;
use Buster;
my @MAUVE_PARAMS =
  qw(output output-alignment output-guide-tree island-size island-output backbone-size max-backbone-gap backbone-output);

#Using attributes with '-' cause problems with function calls so we convert them all to '_'
my %PARAM_CONVERT;
foreach (@MAUVE_PARAMS) {
	my $translate = $_;
	$translate =~ s/-/_/g;
	$PARAM_CONVERT{$_} = $translate;
}

#Other atrributes that can be used
my @ATTRIB = qw(quiet buster buster_node);
my @FILES  =
  qw(output output-alignment output-guide-tree island-output backbone-output);

# The constructor for the class
sub new {
	my ( $class, %arg ) = @_;

	#Bless an anonymous empty hash
	my $self = bless {}, $class;

	#Fill the attributes with mauve parameters
	foreach (@MAUVE_PARAMS) {
		$self->{ $PARAM_CONVERT{$_} } = undef;
	}

	#Fill the other attributes
	foreach (@ATTRIB) {
		$self->{$_} = undef;
	}
	$self->island_size(8000);

	#Set each attribute that is given as an arguement
	foreach ( keys(%arg) ) {
		$self->$_( $arg{$_} );
	}
	return $self;
}

sub run {
	my ( $self, @seqs ) = @_;

	#use the file name of the first sequnce the tmp directory name
	my $tmp_dir;
	if ( $seqs[0] =~ /(\/.+\/)?(\w+)/ ) {
		$tmp_dir = $2;
	} else {
		$tmp_dir = 'mauve_tmp';
	}

	#store the names of all temp files so we can remove them after
	my @remove;

	#Fill parameters that have not been set
	foreach (@FILES) {
		my $param = $PARAM_CONVERT{$_};
		my $tmp   = $self->_get_filename($tmp_dir);
		push( @remove, $tmp );
		$self->$param($tmp);
	}
	my @params;
	foreach (@MAUVE_PARAMS) {
		my $stored_param = $PARAM_CONVERT{$_};
		push( @params, join( "", "--", $_, "=", $self->$stored_param() ) )
		  if defined( $self->$stored_param() );
	}
	my $param_str = join( " ", @params );
	croak "Mauve needs at least 2 sequences to align." if ( scalar(@seqs) < 2 );
	my @seq_sml;
	foreach (@seqs) {
		croak "Can't find sequence file: '$_'" unless -e $_;
		my $tmp = $self->_get_filename($tmp_dir);
		push( @remove, $tmp );
		push( @seq_sml, join( " ", $_, $tmp ) );
	}
	my $seq_str = join( " ", @seq_sml );
	my $command = "mauveAligner $param_str $seq_str";

	#pipe the stderr from the command to a temp file
	my $tmp_stderr = $self->_get_filename($tmp_dir);
	push( @remove, $tmp_stderr );
	$command .= " 2>$tmp_stderr";

	#If Buster flag is set, add "clsh -q node#" to the front of the command
	if ( $self->buster ) {
		my $free_node;

		#if the calling program provides the node to use than use that node
		if ( $self->buster_node ) {
			$free_node = $self->buster_node;

			#otherwise get a free node
		} else {
			my $buster_obj = new Buster();
			$free_node = $buster_obj->get_node();
		}
		$command = "clsh -q $free_node " . $command;
	}

	#run the actual command
	unless ( open( MAUVE, "$command |" ) ) {
		croak "Cannot run $command";
	}

	#Waits until the system call is done before saving to the array
	my @stdout = <MAUVE>;
	close(MAUVE);

	#Open the file that contains the std error from the command call
	open( MAUVE_ERROR, $tmp_stderr )
	  || die "Can't open std_error file: $tmp_stderr \n $!";
	my @stderr = <MAUVE_ERROR>;
	close(MAUVE_ERROR);

	#remove the tmp std error file
	#unlink $tmp_stderr;
	#print any errors from the command to stderr unless the quiet flag is set
	unless ( $self->quiet ) {
		print STDERR @stderr;
	}
	push( @stderr, "\nCommand: $command\n" );

	#store the results in memory as a mauve results object
	my $mauve_results_obj = $self->_save_results( \@stdout, \@stderr );

	#remove all temp files
	$self->_remove_files(@remove);

	#return the mauve results object
	return $mauve_results_obj;
}

sub _save_results {
	my ( $self, $stdout, $stderr ) = @_;

	#Create a new Muave_Results object
	my $mauve_results_obj =
	  new Mauve_Results( 'stdout' => $stdout, 'stderr' => $stderr );
	foreach (@FILES) {
		my $param     = $PARAM_CONVERT{$_};
		my $file_name = $self->$param();
		if ( open( TEMPFILE, $file_name ) ) {
			my @file = <TEMPFILE>;
			$mauve_results_obj->$param( \@file );
			close(TEMPFILE);
		} else {
			carp
"Can't open file: $file_name for mauve param: $param \nOffending Command std err: @$stderr\n";
			last;
		}
	}
	return $mauve_results_obj;
}

sub _get_filename {
	my ( $self, $dir ) = @_;
	my $tmp_dir = "/home/mlangill/tmp/$dir";
	unless ( -e $tmp_dir && -d $tmp_dir ) {
		mkdir($tmp_dir);
	}
	my $tmp_file = mktemp("$tmp_dir/.mauvetmpXXXXXXXXXXXX");
	return $tmp_file;
}

#removes all the files created by mauve
sub _remove_files {
	my ( $self, @remove ) = @_;
	unlink @remove or carp "Couldn't remove tmp files: @remove $!";
}

# This takes the place of methods to set or get the value of an attribute
sub AUTOLOAD {
	my ( $self, $newvalue ) = @_;

	#get the unknown method call
	my $attr = $AUTOLOAD;

	#Keep only the method name
	$attr =~ s/.*:://;

	#Die if the key does not already exist in the hash
	unless ( exists( $self->{$attr} ) ) {
		croak "No such attribute '$attr' exists in the class ";
	}

	# Turn off strict references to enable "magic" AUTOLOAD speedup
	no strict 'refs';

	# define subroutine
	*{$AUTOLOAD} = sub {
		my ( $self, $newvalue ) = @_;
		$self->{$attr} = $newvalue if defined($newvalue);
		return $self->{$attr};
	};

	# Turn strict references back on
	use strict 'refs';

	#Set the new value for the attribute if available
	$self->{$attr} = $newvalue if defined($newvalue);

	#Always return the current value for the attribute
	return $self->{$attr};
}

# When an object is no longer being used, this will be automatically called
# and will adjust the count of existing objects
sub DESTROY {
}
1;

=head1 NAME

Mauve: Used to run the Mauve alignment tool

=head1 Synopsis
 use Mauve;

 my $mauve_obj = new Mauve();

 my $mauve_results_obj = $mauve_obj->run($seqfile1,$seqfile2);

 my @mauve_alignment=@{$mauve_results_obj->output_alignment()};


=head1 AUTHOR

Morgan Langille

=head1 Date

May 23th, 2006

=cut


