#!/usr/bin/perl
#This is used to produce a positive dataset of genomic islands.
#In other words, it finds regions that are unique within a single genome when compared to multiple related genomes
#This runs mauve across all genomes that have had their genome distances calculated using cvtree
#Uses tables in the microbedb database; including the replicon_distance table which should have been previously filled using the run_cvtree.pl script
use strict;
use warnings;

use lib "perl_lib";
use lib "microbedb/MicrobeDB";
use File::Temp qw/ :mktemp /;

#local modules
use Mauve;
use Buster;
use Genome_Picker;

use Bio::Tools::Run::StandAloneBlast;
use Bio::SearchIO;
use Bio::SeqIO;
use Bio::Seq;

use Blast_hit;
use Blast_Filter;

#from the microbe API
use Search;

#name of folder to place output in
my $root_folder = "./";

#name of folder to place tmp files in
my $tmp_folder = "~/tmp"; 

#version of microbedb to use
my $microbedb_version = 1;

#Only use this if running on Buster (a cluster of machines hosted by the Brinkman Lab)
my $parallel = 0;


#These cutoff variables can be adjusted before starting
#################################################

#Maximum distance needed between query genome and comparative genome (cvtree distance)
#"good cutoff"
our $MAX_CUTOFF = 0.42;

#Minimum distance required between EVERY genome to be used
#Increasing this will result in a set of genomes that are more spread out
#However, increasing this too much will limit the number of query genomes that can be used for GI prediction

#"good cutoff"
our $MIN_CUTOFF = 0.10;

#relaxed
#our $MIN_CUTOFF =0.05;

#genomes that have less than this number of genomes with distance between min and max cutoff will not be run

#"good cutoff"
our $MIN_COMPARE_CUTOFF = 3;

#Setting a maxmimum distance that at least one genome has to be within
#This is to ensure that the POSITIVE DATASET is searching for events on a similar time scale
#Decreasing this will reduce our positive dataset

#"good cutoff"
our $MAX_DIST_SINGLE_CUTOFF = 0.3;

#relaxed
#our $MAX_DIST_SINGLE_CUTOFF = 0.42;

#Setting a minimum distance that at least one genome has to be over
#This is to ensure that the NEGATIVE DATASET is finding old enough conserved regions
#Increasing this will reduce our negative dataset

#"good cutoff"
our $MIN_DIST_SINGLE_CUTOFF = 0.34;

#relaxed
#our $MIN_DIST_SINGLE_CUTOFF = 0.05;

#Minimum island size
our $MIN_GI_SIZE = 8000;

################################################

#Make a search and replicon object to get a list of all chromosomes from a particular version
my $search_obj  = new Search();
my $rep_obj     = new Replicon( version_id => $microbedb_version, rep_type => 'chromosome');
my @result_objs = $search_obj->object_search($rep_obj);



my $total_jobs = scalar(@result_objs);
print "Total replicons to process: $total_jobs\n";

my $picker_obj = new Genome_Picker(
    max_cutoff             => $MAX_CUTOFF,
    min_cutoff             => $MIN_CUTOFF,
    min_compare_cutoff     => $MIN_COMPARE_CUTOFF,
    max_dist_single_cutoff => $MAX_DIST_SINGLE_CUTOFF,
    min_dist_single_cutoff => $MIN_DIST_SINGLE_CUTOFF,
    microbedb_version => $microbedb_version,
);

print "Using the following cutoffs for picking genomes:\n";

#print out the cutoffs for picking genomes
$picker_obj->print_object();

my @prepared_genomes;

#loop through each chromosome and find out what genomes can be used for comparision
foreach my $curr_rep_obj (@result_objs) {
   
    #Find/Calculate the the genomes we should use in our comparisions
    my @good_genomes = $picker_obj->find_comparative_genomes($curr_rep_obj);
     
    if (@good_genomes) {

        #print_rep_info($curr_rep_obj);
        push( @prepared_genomes, [ $curr_rep_obj, \@good_genomes ] );
        #print gi_details($curr_rep_obj)."_)\n";
    }

}

my $num_good_queries = scalar(@prepared_genomes);
print "Number of replicons that passed the screening: $num_good_queries\n";

#Uncomment the following if you only want to see what genomes will be run
#exit;

my $folder = $root_folder . "/positive_GIs";
`mkdir -p $folder`;

my $job_count = 0;

#Make buster object and check all nodes on buster
my $buster;
if ($parallel){
     $buster = new Buster();
}


foreach (@prepared_genomes) {
    my $query_replicon     = $_->[0];
    my @comparison_genomes = @{ $_->[1] };

    $job_count++;
    if($parallel){
        #Get a free node on buster or wait until there is a free one
    my $node   = $buster->get_node();
    
        #Use a fork command here so that it uses more nodes on buster
    if ( my $pid = fork() ) {
    #if(0){
     #child of the fork
     #wait a couple of seconds before moving to the next iteration in the loop
        sleep(2);

    }
    else {

        my @all_unique_regions;
        my $num_genomes = scalar(@comparison_genomes);
        my $name        = $query_replicon->rpv_id;
        print "$job_count Running rpv_id: $name on node: $node\n";
        foreach (@comparison_genomes) {

            #do a pairwise alignment
            my $mauve_results_obj =
              do_mauve_alignment( $node, $query_replicon, $_ );  

            #parse output
            push( @all_unique_regions, parse_output($mauve_results_obj) );
        }

     #Finds unique regions that are unique in all of the pairwise comparisions
        my @unique_regions = overlap( $num_genomes, @all_unique_regions );

        #Get the nucleotide sequences from these coordinates
        my @unique_regions_with_seq =
          get_seqs( $query_replicon, @unique_regions );

        my @filtered_unique_regions;
        if ( scalar(@unique_regions) > 0 ) {

            #Create temp blast db
            my $blast_db =
              create_blast_db( $query_replicon, @comparison_genomes );

            #blast
            @filtered_unique_regions =
              blast_screen( $node, $name, $blast_db, @unique_regions );
        }

        #output islands

        open( my $ISLANDS, '>', "$folder/$name" )
          || die "Can't open $folder/$name for writing: $!";
        if ( scalar(@filtered_unique_regions) > 0 ) {
            foreach my $curr_island (@filtered_unique_regions) {

                #print Dumper(@filtered_unique_regions);
                if ( defined($curr_island) ) {
                    print $ISLANDS "$curr_island->[0]\t$curr_island->[1]\n";
                }
            }
        }
        close($ISLANDS);
        exit;

    }  
    }else{
        my @all_unique_regions;
        my $num_genomes = scalar(@comparison_genomes);
        my $name        = $query_replicon->rpv_id;
        print "$job_count Running rpv_id: $name \n";
        foreach (@comparison_genomes) {

            #do a pairwise alignment
            my $mauve_results_obj =
              do_mauve_alignment( 0, $query_replicon, $_ );  

            #parse output
            push( @all_unique_regions, parse_output($mauve_results_obj) );
        }

     #Finds unique regions that are unique in all of the pairwise comparisions
        my @unique_regions = overlap( $num_genomes, @all_unique_regions );

        #Get the nucleotide sequences from these coordinates
        my @unique_regions_with_seq =
          get_seqs( $query_replicon, @unique_regions );

        my @filtered_unique_regions;
        if ( scalar(@unique_regions) > 0 ) {

            #Create temp blast db
            my $blast_db =
              create_blast_db( $query_replicon, @comparison_genomes );

            #blast
            @filtered_unique_regions =
              blast_screen( 0, $name, $blast_db, @unique_regions );
        }

        #output islands

        open( my $ISLANDS, '>', "$folder/$name" )
          || die "Can't open $folder/$name for writing: $!";
        if ( scalar(@filtered_unique_regions) > 0 ) {
            foreach my $curr_island (@filtered_unique_regions) {

                #print Dumper(@filtered_unique_regions);
                if ( defined($curr_island) ) {
                    print $ISLANDS "$curr_island->[0]\t$curr_island->[1]\n";
                }
            }
        }
        close($ISLANDS);
    }
    

  
}

sub blast_screen {
    my ( $node, $name, $db, @unique_regions ) = @_;

    #get a tmp file to write the coordinates out to
    my $tmp_gi_file = get_tmp_filename("blast_screen");

    open( my $GI, '>', "$tmp_gi_file" )
      || die "Can't open $tmp_gi_file for writing: $!";

    foreach (@unique_regions) {
        print $GI "$_->[0]\t$_->[1]\n";
    }
    close($GI);

    my $command = "./util_scripts/blast_islands.pl $db $name $tmp_gi_file";
    if($node eq 0){
        $command = "clsh -q $node ".$command;
    }
    `$command`;

    open( my $GI2, '<', $tmp_gi_file )
      || die "Can't open $tmp_gi_file for reading: $!";

    my @filtered_gi;
    while ( my $curr_gi = <$GI2> ) {
        if ( $curr_gi =~ /(\d+)\t(\d+)/ ) {
            push( @filtered_gi, [ $1, $2 ] );
        }
        else {
            die "Can't parse filtered gi file: $curr_gi :$!";
        }
    }
    close($GI2);
    `rm $tmp_gi_file`;

    return @filtered_gi;
}

#Create a temporary blast database
sub create_blast_db {

    #shift; #uncomment if reference genome is to be ignored
    my @genomes = @_;
    my @dbseqs;
    foreach (@genomes) {
        push( @dbseqs, $_->get_filename('fna') );
    }
    my $db = get_tmp_filename("blast");
    system("cat @dbseqs > $db");

    system("formatdb -i $db -p F");
    return $db;
}

sub get_tmp_filename {
    my ($dir) = @_;
    my $tmp_dir = "$tmp_folder/$dir";
    unless ( -e $tmp_dir && -d $tmp_dir ) {
        mkdir($tmp_dir);
    }
    my $tmp_file = mktemp("$tmp_dir/XXXXXXXXXXXXXX");
    return $tmp_file;
}

#Get the nucleotide sequences for each putative genomic island
sub get_seqs {
    my ( $rep_obj, @unique_regions ) = @_;
    return unless scalar(@unique_regions) > 0;

    #Create the data structure that we will return
    my @unique_regions_with_seq;

    #Going to do a Search to get the entire replicon sequence from microbedb
    my $search_obj = new Search( return_seqs => 1 );
    my ($complete_rep_obj) = $search_obj->object_search($rep_obj);
    my $rep_seq = $complete_rep_obj->rep_seq();

    #Iterate through each unique region
    foreach my $coordinates (@unique_regions) {

        #Set start and end coordinates of the unique region
        my $start = $coordinates->[0];
        my $end   = $coordinates->[1];

#get the seq. $start -1 to include the $start nucleotide and +1 to the length to include the last nucleotide
        my $seq = substr( $rep_seq, $start - 1, $end - $start + 1 );
        push( @unique_regions_with_seq, [ $start, $end, $seq ] );
    }
    return @unique_regions_with_seq;
}

#Finds unique regions that are unique in all of the pairwise comparisions
#Note: the code in this is badly written, but I never got around to rewriting it!
sub overlap {
    my ( $num_genomes, @all_unique_regions ) = @_;
    my $i = 0;
    my ( @start, @end );
    foreach (@all_unique_regions) {
        ( $start[$i], $end[$i] ) = ( $_->[0], $_->[1] );
        $i++;
    }
    my ( $found, $temp_end, @finalstart, @finalend );
    foreach my $temp_start (@start) {
        $found    = 0;
        $temp_end = undef;
        for ( my $j = 0; $j < @start; $j++ ) {
            if ( $temp_start >= $start[$j] && $temp_start < $end[$j] ) {
                $found++;

                #Keeps track of the nearest end point of overlapping gaps
                if ( !( defined($temp_end) ) || $end[$j] < $temp_end ) {
                    $temp_end = $end[$j];
                }

                #found gap region that is present across all genomes
                if ( $found >= $num_genomes ) {

                    #Check if region was already found
                    my $inlist = 0;
                    for ( my $k = 0; $k < @finalstart; $k++ ) {
                        if (   $finalstart[$k] == $temp_start
                            && $finalend[$k] == $temp_end )
                        {
                            $inlist = 1
                              ; #set the flag to indicate that the gap is already in the list
                            last;    #exit the for loop
                        }
                    }

                    #if not in list already add it to the list
                    if ( !($inlist) ) {
                        push( @finalstart, $temp_start );
                        push( @finalend,   $temp_end );
                    }
                    last
                      ; # exits the for loop once overlap is found in all alignments
                }
            }
        }
    }
    my @unique_regions;
    for ( my $i = 0; $i < @finalstart; $i++ ) {
        push( @unique_regions, [ $finalstart[$i], $finalend[$i] ] );
    }
    return @unique_regions;
}

sub parse_output {
    my ($mauve_results_obj) = @_;
    my @unique_regions;

    #parse xmfa
    my @xmfa_output = @{ $mauve_results_obj->output_alignment() };
    my $last_number = 0;
    my $gap_area    = 0;

    #start out not in gapped section
    foreach (@xmfa_output) {
        if (/\> 1:(\d*)-(\d*)/) {
            my $start = $1;
            my $end   = $2;
            my $size  = $end - $start + 1;
            if ( !$gap_area ) {
                if ( $last_number > $end ) {
                    $gap_area = 1;

                    #found a gap. Check size and put in array
                    if ( $size >= $MIN_GI_SIZE ) {
                        push( @unique_regions, [ $start, $end ] );
                    }
                }
                else {
                    $last_number = $end;
                }
            }
            else {
                if ( $size >= $MIN_GI_SIZE ) {
                    push( @unique_regions, [ $start, $end ] );
                }
            }
        }
    }
    my @island_output = @{ $mauve_results_obj->island_output() };
    foreach (@island_output) {
        if (/\d\t(-?\d+)\t(-?\d+)\t\d\t(-?\d+)\t(-?\d+)/) {
            if ( abs( $2 - $1 ) > abs( $4 - $3 ) ) {
                push( @unique_regions, [ $1, $2 ] );
            }
        }
        else {
            warn "Problem parsing mauve island output file: $!";
        }
    }
    return @unique_regions;
}

#Calls the Mauve class to do the actual mauve alignment
#Input: a node on buster, and an array of Replicon objects
sub do_mauve_alignment {
    my ( $node, @genomes ) = @_;

    #Create a Mauve object
    my $mauve_obj;
    if($node eq 0){
    
        $mauve_obj = new Mauve(
        buster      => 0,
        'quiet'     => 1,
        island_size => $MIN_GI_SIZE
    );
    }else{
      #Create a Mauve object
        $mauve_obj = new Mauve(
        buster      => 1,
        buster_node => $node,
        'quiet'     => 1,
        island_size => $MIN_GI_SIZE
    );  
    }
   
    my @seqs;
    my $search_obj = new Search( return_obj => 'GenomeProject' );
    foreach my $curr_genome (@genomes) {

        #need the paths to the genome projects
        my ($gpo) = $search_obj->object_search($curr_genome);

        #produce the path to the proper file
        my $seq = $gpo->gpv_directory() . $curr_genome->file_name() . ".fna";
        push( @seqs, $seq );
    }

    #print "Running: $seqs[0] \n";
    if ( scalar(@seqs) > 2 ) {
        warn "Warning More than 2 sequences: @seqs\n";
    }

    return $mauve_obj->run(@seqs);
}



sub gi_details {
    my $replicon = shift;
    my $so       = new Search();    
    $so->return_obj('GenomeProject');
    my ($go) = $so->object_search($replicon);
    my $patho = $go->patho_status();
    
    if($patho eq 'unknown'){
        $patho = 'pathogen';
    }
    
    return $patho;

}

