#!/usr/bin/perl 

#Updated/Edited by: Morgan

# AUTH:	Anastasia A.  Fedynak
# DATE: 22 March, 2004B
# DESC: Blast the given .islands file (file containing putative islands)
#	against reference fasta sequences.  Filter out regions with large
#	blast hits, and write out the island sequences and their corresponding
#	open reading frames.
# FILE: blaster.pl
# USAGE:
#	perl blaster.pl [options] <file.islands> <file1.fasta> .. <file3.fasta>

use Bio::Tools::Run::StandAloneBlast;
use Bio::SearchIO;
use Bio::SeqIO;
use Bio::Seq;


use strict;
use warnings;

use lib "/home/mlangill/common/workspace/GI_Finder/perl_lib";
use lib "/home/mlangill/common/workspace/microbedb/MicrobeDB";
use Search;
use Blast_hit;
use Blast_Filter;


my $program = "blastn";		# default blast program
my $e_cutoff = "1";		# default e-value cutoff
my $min_island_size = 8000;	# default minimum island size


my $db = $ARGV[0];
my $rpv_id = $ARGV[1];
my $tmp_gi_file = $ARGV[2];


my @unfiltered_islands = read_gi_file($tmp_gi_file);
@unfiltered_islands = get_seqs($rpv_id,@unfiltered_islands);

my @all_islands;
if(scalar(@unfiltered_islands) > 0){
	@all_islands = blast_sequences($db,@unfiltered_islands);
}else{
	warn "No islands to blast against: $!";
}

write_gi_file($tmp_gi_file,@all_islands);

print "Done\n";

sub write_gi_file {
	my($tmp_gi_file,@unique_regions) =@_;
	open(GI,">$tmp_gi_file") || die "Can't open $tmp_gi_file for writing: $!";
	
	foreach(@unique_regions){
		my $start = $_->{begin} + $_->{offset};
		my $end = $_->{end} + $_->{offset};
		print GI "$start\t$end\n";	
	}
	close(GI);	
}

sub read_gi_file {
	my ($tmp_gi_file)=@_;
	open(GI2,$tmp_gi_file) || die "Can't open $tmp_gi_file for reading: $!";
	
	my @gi;
	while(my $curr_gi = <GI2>){
		if($curr_gi =~ /(\d+)\t(\d+)/){
			push(@gi, [$1,$2]);
		}else{
			die"Can't parse filtered gi file: $curr_gi :$!";	
		}
	}
	close(GI2);
	return @gi;
}

#Get the nucleotide sequences for each putative genomic island
sub get_seqs {
	my ( $rpv_id, @unique_regions ) = @_;
	return undef unless scalar(@unique_regions) > 0;
	
	my $rep_obj = new Replicon(rpv_id => $rpv_id);
	
	#Create the data structure that we will return
	my @unique_regions_with_seq;

	#Going to do a Search to get the entire replicon sequence from microbedb
	my $search_obj = new Search( return_seqs => 1 );
	my ($complete_rep_obj) = $search_obj->object_search($rep_obj);
	my $rep_seq = $complete_rep_obj->rep_seq();

	#Iterate through each unique region
	foreach my $coordinates (@unique_regions) {

		#Set start and end coordinates of the unique region
		my $start = $coordinates->[0];
		my $end   = $coordinates->[1];

#get the seq. $start -1 to include the $start nucleotide and +1 to the length to include the last nucleotide
		my $seq = substr( $rep_seq, $start - 1, $end - $start + 1 );
		push( @unique_regions_with_seq, [ $start, $end, $seq ] );
	}
	return @unique_regions_with_seq;
}


# blast_sequences(): 
# ARGS: None
# RETU: @all_islands	Coordinates of islands
# DESC: blasts sequences in $islands_file against database.  For each sequence
#	filter out the blast hits and store the remaining island in '@all_islands'.
sub blast_sequences {
	my ($db,@unfiltered_islands) = @_;
	
	my $e_cutoff = 1;
	
	foreach(@unfiltered_islands){
		
		my $offset = $_->[0];
		my $end = $_->[1];
		my $nuc = $_->[2];
		my $seq = new Bio::Seq(-display_id =>"$offset,$end", -seq => $nuc);
		
		my $island_size = $seq->length;
		
		my $isl_name ="";

		my @params = ('program' => 'blastn', 'database' => $db,
		  'e' => $e_cutoff);

		my $blast = Bio::Tools::Run::StandAloneBlast->new(@params);
		my $blast_report = $blast->blastall($seq);
		while ( my $result = $blast_report->next_result() ) { # blast hits for each island
			my $num_hits = 0;
			my @islands =();
			my @blast_hits =();
			while( my $hit = $result->next_hit) {
				while (my $hsp = $hit->next_hsp) {
					my $bhit = createNewBlastHit($hit->name,$hsp);
					push (@blast_hits, $bhit);
					$num_hits++;
				}
			}
			if ($num_hits == 0) { 
				@islands = getIslandCoords($offset, $isl_name, $island_size, ());
			}else {
				@blast_hits = blast_filter($island_size, @blast_hits);
				@islands = getIslandCoords($offset, $isl_name, $island_size, @blast_hits);
				
			}
		push (@all_islands, @islands);
		}
	}

	#system("rm $db*");

	return (@all_islands);
}


# createNewBlastHit(): 
# ARGS: $name	Name of the organism
#	$hsp	Bioperl hsp object
# RETU: $bit	A new Blast_hit object
# DESC: Stores information of a single blast hit and returns it.
sub createNewBlastHit {
	(my $name, my $hsp) = @_;
	my $bhit = new Blast_hit;

	$bhit->start($hsp->query->start);
	$bhit->end($hsp->query->end);
	$bhit->length($hsp->length);
	$bhit->score($hsp->score);
	$bhit->frac_id($hsp->frac_identical);
	$bhit->name($name);

	return $bhit;
}
	
	
# getIslandCoords(): 
# ARGS: $offset		Offset of the island in the organism nucleotide sequence
#	$isl_name	Name of the organism
#	$isl_size	Size of the island
#	@b_hits		Array containing all Blast_hit objects
# RETU: @island_coords	Array containing all new island coordinates
# DESC: Calculate coordinates of islands after removing portions of the island
#	that have a blast hit against
sub getIslandCoords {

	my ($offset, $isl_name, $isl_size, @b_hits) = @_;
	my $begin = 0;
	my $end = undef;
	my $diff = undef;
	my @island_coords = ();

	my $num_hits = scalar(@b_hits);
	

	for (my $i = 0; $i <= $num_hits; $i++) {	
		my $hit = $b_hits[$i];
		if ($i == $num_hits) {
			$end = $isl_size;
		}
		else {
			$end = $hit->start;
		}

		$diff = $end - $begin;
		if ($diff > $min_island_size) {
			my $seq_start = $begin + $offset;
			my $seq_end = $end + $offset;
			my %isl = ();
			$isl{begin} = $begin;
			$isl{end} = $end;
			$isl{offset} = $offset;
			$isl{len} = $diff;
			$isl{name} = $isl_name;
			push (@island_coords, \%isl);
			
		}
		if ($i !=$num_hits) {
			$begin = $hit->end;
		}
	}
	return @island_coords;
}





